# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the configuration variables and implements their
initialisation, loading, and saving in a file.

Example usage:
import config
...
# XXX We shouldn't have the default values contain localised parts:
self.config = config.Config(_('Default Profile'), _("by") + " %A " +\
        _("from") + " %B")
"""

import logging
import os
import hashlib
from configparser import RawConfigParser

from sonata import misc, consts
from sonata.song import SongRecord

# Constant to express a None value
LIB_NODATA = "!NONE!"

logger = logging.getLogger(__name__)


class ConfigParser(RawConfigParser):
    """Override the default configuration parser with specific loader"""

    def getlist(self, section, name, sep=","):
        """Get a list of values, separated by `sep`"""
        value = self.get(section, name)
        return [v.strip() for v in value.split(sep)]

    def getlistint(self, section, name, sep=","):
        """Load a list of integer values, separated by `sep`."""
        return [int(v) for v in self.getlist(section, name, sep)]


class Serializer:
    """Helper to serialize specific values.

    Those methods can be function, but it makes easier to locate and load them
    if they are stored here."""

    @staticmethod
    def list(value, sep=","):
        return ",".join(str(s) for s in value)
    listint = list


class Config:
    """This class contains the configuration variables as attributes.

    Each new variable should be initialised to its default value
    in __init__, loaded from a file in settings_load_real, and
    saved to a file in settings_save_real.

    XXX This is mostly ConfigParser plus some custom serialization work.
    """

    CONFIG_PATH = os.path.expanduser('~/.config/sonata/sonatarc')

    def __init__(self, default_profile_name, currsongformat2):
        self.wd = SongRecord(path="/")

        self._options = {
            'audioscrobbler': {
                'as_enabled': ('use_audioscrobbler', 'boolean', False),
                'as_password_md5': ('password_md5', '', ''),
                'as_username': ('username', '', '')},
            'connection': {
                'autoconnect': ('auto', 'boolean', True),
                'profile_num': ('profile_num', 'int', 0)},
            'currformat': {
                'currentformat': ('current', '', '%A - %T|%L'),
                'currsongformat1': ('currsong1', '', '%T'),
                'currsongformat2': ('currsong2', '', currsongformat2),
                'libraryformat': ('library', '', '%A - %T'),
                'titleformat': ('title', '', '[Sonata] %A - %T')},
            'library': {
                'lib_view': ('lib_view', 'int', consts.VIEW_FILESYSTEM)},
            'notebook': {
                'current_tab_pos': ('current_tab_pos', 'int', 0),
                'current_tab_visible': ('current_tab_visible', 'boolean', True),
                'info_tab_pos': ('info_tab_pos', 'int', 4),
                'info_tab_visible': ('info_tab_visible', 'boolean', True),
                'library_tab_pos': ('library_tab_pos', 'int', 1),
                'library_tab_visible': ('library_tab_visible', 'boolean', True),
                'playlists_tab_pos': ('playlists_tab_pos', 'int', 2),
                'playlists_tab_visible': ('playlists_tab_visible', 'boolean', True),
                'streams_tab_pos': ('streams_tab_pos', 'int', 3),
                'streams_tab_visible': ('streams_tab_visible', 'boolean', True)},
            'player': {
                'art_location': ('art_location', 'int', consts.ART_LOCATION_HOMECOVERS),
                'art_location_custom_filename': ('art_location_custom_filename', '', ''),
                'columnwidths': ('columnwidths', 'listint', [325, 10]),
                'covers_pref': ('covers_pref', 'int', consts.ART_LOCAL_REMOTE),
                'covers_type': ('covers_type', 'int', 1),
                'decorated': ('decorated', 'boolean', True),
                'existing_playlist_option': ('existing_playlist', 'int', 0),
                'expanded': ('expanded', 'boolean', True),
                'h': ('h', 'int', 300),
                'info_album_expanded': ('info_album_expanded', 'boolean', True),
                'info_art_enlarged': ('info_art_enlarged', 'boolean', False),
                'info_lyrics_expanded': ('info_lyrics_expanded', 'boolean', True),
                'info_song_expanded': ('info_song_expanded', 'boolean', True),
                'info_song_more': ('info_song_more', 'boolean', False),
                'infofile_path': ('infofile_path', '', '/tmp/xmms-info'),
                'initial_run': ('initial_run', 'boolean', True),
                'last_search_num': ('search_num', 'int', 0),
                'lyrics_location': ('lyrics_location', 'int', consts.LYRICS_LOCATION_HOME),
                'minimize_to_systray': ('minimize', 'boolean', False),
                'ontop': ('ontop', 'boolean', False),
                'popup_option': ('popup_time', 'int', 2),
                'screen': ('screen', 'int', 0),
                'show_covers': ('covers', 'boolean', True),
                'show_header': ('show_header', 'boolean', True),
                'show_lyrics': ('lyrics', 'boolean', True),
                'show_notification': ('notification', 'boolean', False),
                'show_playback': ('playback', 'boolean', True),
                'show_progress': ('progressbar', 'boolean', True),
                'show_statusbar': ('statusbar', 'boolean', False),
                'show_trayicon': ('trayicon', 'boolean', True),
                'sticky': ('sticky', 'boolean', False),
                'stop_on_exit': ('stop_on_exit', 'boolean', False),
                'tabs_expanded': ('tabs_expanded', 'boolean', False),
                'traytips_notifications_location': ('notif_location', 'int', 0),
                'update_on_start': ('update_on_start', 'boolean', False),
                'url_browser': ('browser', '', ''),
                'use_infofile': ('use_infofile', 'boolean', False),
                'w': ('w', 'int', 400),
                'withdrawn': ('withdrawn', 'boolean', False),
                'x': ('x', 'int', 0),
                'xfade': ('xfade', 'int', 0),
                'xfade_enabled': ('xfade_enabled', 'boolean', False),
                'y': ('y', 'int', 0)},
            'plugins': {
                'autostart_plugins': ('autostart_plugins', 'list', []),
                'known_plugins': ('known_plugins', 'list', [])},
            'tags': {
                'tags_use_mpdpath': ('use_mpdpaths', 'boolean', False)}
        }

        self._indexed_options = {
            'streams': ('num_streams', {
                'stream_names': ('names', '', []),
                'stream_uris': ('uris', '', []),
            }),
            'profiles': ('num_profiles', {
                'profile_names': ('names', '', [default_profile_name]),
                'musicdir': ('musicdirs', '', ["~/music"]),
                'host': ('hosts', '', ['localhost']),
                'port': ('ports', 'int', [6600]),
                'password': ('passwords', '', ['']),
            })
        }

    @property
    def current_musicdir(self):
        return self.musicdir[self.profile_num]

    def settings_load_real(self):
        """Load configuration from file"""

        conf = ConfigParser()
        misc.create_dir(os.path.dirname(self.CONFIG_PATH))
        conf.read(self.CONFIG_PATH)

        # Load all the "simple" options, as described in self._options, and set
        # them as instance attribute.
        for section, attributes in self._options.items():
            for attribute, (opt_key, type, default) in attributes.items():
                if conf.has_option(section, opt_key):
                    try:
                        value = getattr(conf, 'get' + type)(section, opt_key)
                    except Exception as e:
                        # BBB: we need to expect some errors since Sonata uses
                        # to write None values for "int"-type settings, which
                        # fail to be loaded when using getint(). The new code
                        # should write better values each time. Consider
                        # removing this try/except clause when configuration
                        # files are "clean".
                        value = default
                        # This should be safe in all cases
                        faulty_value = conf.get(section, opt_key)
                        logger.warning(
                            "Can't load %r from section %r (as %s). Value is %r",
                            opt_key, section, type if type else "str",
                            faulty_value)
                else:
                    value = default
                setattr(self, attribute, value)

        # Load all the attributes which have several values and are indexed.
        for section, (index_name, attributes) in self._indexed_options.items():
            if not conf.has_option(section, index_name):
                num = 0
            else:
                num = conf.getint(section, index_name)

            for attribute, (key, type, default) in attributes.items():
                if num == 0:
                    setattr(self, attribute, default)
                else:
                    setattr(self, attribute, [])

                for i in range(num):
                    opt_key = "%s[%d]" % (key, i)
                    value = getattr(conf, 'get' + type)(section, opt_key)
                    getattr(self, attribute).append(value)

        # Finally, load attributes related to the library. This is a bit weird
        # so we use an helper function to make it easier:
        def lib_get(name):
            # Helper function to load attributes related to the library.
            value = None
            if conf.has_option('library', name):
                value = conf.get('library', name)
            if value == LIB_NODATA:
                value = None
            return value

        if conf.has_section('library'):
            album  = lib_get('lib_album')
            artist = lib_get('lib_artist')
            genre  = lib_get('lib_genre')
            year   = lib_get('lib_year')
            path   = lib_get('lib_path')
            self.wd = SongRecord(album, artist, genre, year, path)

        # Finally, patch some values:
        self.musicdir = [misc.sanitize_musicdir(v) for v in self.musicdir]
        # Ensure we have a valid profile number:
        self.profile_num = max(0, min(self.profile_num,
                                      len(self.profile_names) - 1))

        # Specifying remote artwork first is too confusing and probably
        # rarely used, so we're removing this option and defaulting users
        # back to the default 'local, then remote' option.
        # Backward compatibility
        if self.covers_pref > consts.ART_LOCAL_REMOTE:
            self.covers_pref = consts.ART_LOCAL_REMOTE

    def settings_save_real(self):
        """Save configuration in file"""

        conf = ConfigParser()

        # First, write all the "simple" attributes in their respective section:
        for section, attributes in self._options.items():
            if not conf.has_section(section):
                conf.add_section(section)

            for attribute, (name, type, default) in attributes.items():
                value = getattr(self, attribute)
                if hasattr(Serializer, type):
                    value = getattr(Serializer, type)(value)
                conf.set(section, name, value)

        # Then, write all the attributes which are grouped together, and index
        # them.
        for section, (index_name, attributes) in self._indexed_options.items():
            if not conf.has_section(section):
                conf.add_section(section)
            value_index = 0

            for attribute, (name, type, default) in attributes.items():
                for i, value in enumerate(getattr(self, attribute)):
                    conf.set(section, "%s[%d]" % (name, i), value)
                    value_index = i + 1

            conf.set(section, index_name, value_index)

        # Finally, save the settings related to the library. Again, we use an
        # helper function to help us to do the job:
        def lib_set(name):
            value = getattr(self.wd, name)
            conf.set('library', 'lib_' + name,
                     value if value is not None else LIB_NODATA)

        if not conf.has_section('library'):
            conf.add_section('library')
        lib_set('album')
        lib_set('artist')
        lib_set('genre')
        lib_set('year')
        lib_set('path')

        # Now, we can properly save the configuration file:
        try:
            with open(self.CONFIG_PATH, 'w', encoding="utf-8") as rc:
                conf.write(rc)
        except IOError as e:
            self.logger.warning("Couldn't write configuration into %r: %s",
                                self.CONFIG_PATH, e)



#!/usr/bin/env python3
# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""Sonata is a simple GTK+ client for the Music Player Daemon."""

import sys
if sys.version_info <= (3, 3):
    sys.stderr.write("Sonata requires Python 3.3+\n")
    sys.exit(1)

import gettext
import locale
import logging
import os
import platform
import threading  # needed for interactive shell
import gi

def run():
    """Main entry point of Sonata"""

    # TODO: allow to exit the application with Ctrl+C from the terminal
    # This is a fix for https://bugzilla.gnome.org/show_bug.cgi?id=622084
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # XXX insert the correct sonata package dir in sys.path

    logging.basicConfig(
        level=logging.WARNING,
        format="[%(asctime)s] [%(threadName)s] %(name)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr)

    logger = logging.getLogger(__name__)

    try:
        import sonata
    except ImportError:
        logger.critical("Python failed to find the sonata modules.")
        logger.critical("Searched in the following directories:\n%s",
                        "\n".join(sys.path))
        logger.critical("Perhaps Sonata is improperly installed?")
        sys.exit(1)

    try:
        from sonata.version import version
    except ImportError:
        logger.critical("Python failed to find the sonata modules.")
        logger.critical("An old or incomplete installation was "
                        "found in the following directory: %s",
                        os.path.dirname(sonata.__file__))
        logger.critical("Perhaps you want to delete it?")
        sys.exit(1)

    # XXX check that version.VERSION is what this script was installed for

    ## Apply global fixes:

    if platform.system() == 'Linux':
        sys.argv[0] = "sonata"
        import ctypes
        libc = ctypes.CDLL('libc.so.6')
        PR_SET_NAME = 15
        libc.prctl(PR_SET_NAME, b"sonata", 0, 0, 0)

    ## Apply locale and translation:
    # Try to find a "good" locale directory.
    for path in [
        # This is useful when working from the source repository
        os.path.join(os.path.dirname(sonata.__file__), "share", "locale"),
        # This is useful when Sonata is installed in a special place
        os.path.join(sonata.__file__.split('/lib')[0], 'share', 'locale'),
    ]:
        if os.path.exists(path):
            locales_path = path
            break
    else:
        # This tells gettext to look at the default place for the translation
        # files.
        locales_path = None

    # Gtk.Builder uses gettext functions from C library. Enable
    # correct localization for these functions with the locale
    # module. See:
    # https://docs.python.org/3/library/locale.html#access-to-message-catalogs
    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error as e:
        # If locale is not supported by C library, the initial call to
        # locale.setlocale will fail and raise an exception. Any Glade
        # strings would not be translated. But native python strings
        # would still be translated if the .mo files are included in
        # sonata!
        #
        # To prevent a mix of languages, disable translation for both:
        # 1. explicitly set locale to 'C' (default strings)
        # 2. don't provide python gettext with any translatable
        #    strings (localedir=None), but still install the required
        #    _() function
        logger.error("setlocale() failed: %s. Falling back to default locale.", e)
        locale.setlocale(locale.LC_ALL, 'C')
        gettext.install(True, localedir=None, names=["ngettext"])
    else:
        gettext.install('sonata', locales_path, names=["ngettext"])

        # bindtextdomain() is GNU libc specific and may not be available
        # on other systems (e.g. OSX)
        if hasattr(locale, 'bindtextdomain'):
            locale.bindtextdomain('sonata', locales_path)

        gettext.textdomain('sonata')
        gettext.bindtextdomain('sonata', locales_path)


    ## Check initial dependencies:
    try:
        import mpd
    except:
        logger.critical("Sonata requires python-mpd2. Aborting...")
        sys.exit(1)


    ## Initialize the plugin system:

    from sonata.pluginsystem import pluginsystem
    pluginsystem.find_plugins()
    pluginsystem.notify_of('enablables',
                   lambda plugin, cb: cb(True),
                   lambda plugin, cb: cb(False))


    ## Load the command line interface:

    from sonata import cli
    args = cli.Args()
    args.parse(sys.argv)

    ## Deal with GTK:

    if not args.skip_gui:
        # importing gtk does sys.setdefaultencoding("utf-8"), sets locale etc.
        gi.require_version('Gdk', '3.0')
        gi.require_version('GdkPixbuf', '2.0')
        gi.require_version('Gtk', '3.0')
        gi.require_version('Pango', '1.0')
        from gi.repository import Gtk, Gdk
    else:
        class FakeModule:
            pass
        # make sure the ui modules aren't imported
        for m in 'gtk', 'pango', 'sonata.ui', 'sonata.breadcrumbs':
            if m in sys.modules:
                logger.warning(
                    "Module %s imported in CLI mode (it should not)", m)
            else:
                sys.modules[m] = FakeModule()


    ## Global init:

    from socket import setdefaulttimeout as socketsettimeout
    socketsettimeout(5)

    if not args.skip_gui:
        Gdk.threads_init()

    ## CLI actions:

    args.execute_cmds()

    ## Load the main application:

    from sonata import main


    def on_application_activate(application):
        windows = application.get_windows()

        if windows:
            for window in windows:
                window.present()
        else:
            sonata = main.Base(args)
            sonata.window.set_application(application)

    app = Gtk.Application(application_id="org.MPD.Sonata")
    app.connect("activate", on_application_activate)

    ## Load the shell
    # yo dawg, I heard you like python,
    # so I put a python shell in your python application
    # so you can debug while you run it.
    if args.start_shell:
        # the enviroment used for the shell
        scope = dict(list(globals().items()) + list(locals().items()))
        def run_shell():
            try:
                import IPython
                IPython.embed(user_ns=scope)
            except ImportError as e: # fallback if ipython is not avaible
                import code
                shell = code.InteractiveConsole(scope)
                shell.interact()
            # quit program if shell is closed,
            # This is the only way to close the program clone in this mode,
            # because we can't close the shell thread easily
            from gi.repository import Gtk
            Gtk.main_quit()
        threading.Thread(target=run_shell, name="Shell").start()

    try:
        app.run([])
    except KeyboardInterrupt:
        Gtk.main_quit()



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""
This module makes Sonata submit the songs played to a Last.fm account.

Example usage:
import scrobbler
self.scrobbler = scrobbler.Scrobbler(self.config)
self.scrobbler.import_module()
self.scrobbler.init()
...
self.scrobbler.handle_change_status(False, self.prevsonginfo)
"""

import logging
import os
import sys
import threading # init, np, post start threads init_thread, do_np, do_post
import time

audioscrobbler = None # imported when first needed

from sonata import mpdhelper as mpdh


class Scrobbler:

    def __init__(self, config):
        self.logger = logging.getLogger(__name__)
        self.config = config

        self.scrob = None
        self.scrob_post = None
        self.scrob_start_time = ""
        self.scrob_playing_duration = 0
        self.scrob_last_prepared = ""

        self.elapsed_now = None

    def import_module(self, _show_error=False):
        """Import the audioscrobbler module"""
        # We need to try to import audioscrobbler either when the app starts
        # (if as_enabled=True) or if the user enables it in prefs.
        global audioscrobbler
        if audioscrobbler is None:
            from sonata import audioscrobbler

    def imported(self):
        """Return True if the audioscrobbler module has been imported"""
        return audioscrobbler is not None

    def init(self):
        """Initialize the Audioscrobbler support if enabled and configured"""
        if audioscrobbler is not None and self.config.as_enabled and \
           len(self.config.as_username) > 0 and \
           len(self.config.as_password_md5) > 0:
            thread = threading.Thread(target=self.init_thread)
            thread.name = "Scrobbler"
            thread.daemon = True
            thread.start()

    def init_thread(self):
        if self.scrob is None:
            self.scrob = audioscrobbler.AudioScrobbler()
        if self.scrob_post is None:
            self.scrob_post = self.scrob.post(self.config.as_username,
                                              self.config.as_password_md5,
                                              verbose=True)
        else:
            if self.scrob_post.authenticated:
                return # We are authenticated
            else:
                self.scrob_post = self.scrob.post(self.config.as_username,
                                                  self.config.as_password_md5,
                                                  verbose=True)
        try:
            self.scrob_post.auth()
        except Exception as e:
            self.logger.error("Error authenticating audioscrobbler: %r", e)
            self.scrob_post = None
        if self.scrob_post:
            self.retrieve_cache()

    def handle_change_status(self, state, prevstate, prevsonginfo,
                             songinfo=None, mpd_time_now=None):
        """Handle changes to play status, submitting info as appropriate"""
        if prevsonginfo and 'time' in prevsonginfo:
            prevsong_time = prevsonginfo.time
        else:
            prevsong_time = None

        if state in ('play', 'pause'):
            elapsed_prev = self.elapsed_now
            self.elapsed_now, length = [float(c) for c in
                                        mpd_time_now.split(':')]
            current_file = songinfo.file
            if prevstate == 'stop':
                # Switched from stop to play, prepare current track:
                self.prepare(songinfo)
            elif (prevsong_time and
                  (self.scrob_last_prepared != current_file or
                   (self.scrob_last_prepared == current_file and
                    elapsed_prev and self.elapsed_now <= 1 and
                    self.elapsed_now < elapsed_prev and length > 0))):
                # New song is playing, post previous track if time criteria is
                # met. In order to account for the situation where the same
                # song is played twice in a row, we will check if previous
                # elapsed time was larger than current and we're at the
                # beginning of the same song now
                if self.scrob_playing_duration > 4 * 60 or \
                   self.scrob_playing_duration > int(prevsong_time) / 2:
                    if self.scrob_start_time != "":
                        self.post(prevsonginfo)
                # Prepare current track:
                self.prepare(songinfo)
            # Keep track of the total amount of time that the current song
            # has been playing:
            now = time.time()
            if prevstate != 'pause':
                self.scrob_playing_duration += now - self.scrob_prev_time
            self.scrob_prev_time = now
        else: # stopped:
            self.elapsed_now = 0
            if prevsong_time:
                if self.scrob_playing_duration > 4 * 60 or \
                   self.scrob_playing_duration > int(prevsong_time) / 2:
                    # User stopped the client, post previous track if time
                    # criteria is met:
                    if self.scrob_start_time != "":
                        self.post(prevsonginfo)

    def auth_changed(self):
        """Try to re-authenticate"""
        if self.scrob_post:
            if self.scrob_post.authenticated:
                self.scrob_post = None

    def prepare(self, songinfo):
        if audioscrobbler is not None:
            self.scrob_start_time = ""
            self.scrob_last_prepared = ""
            self.scrob_playing_duration = 0
            self.scrob_prev_time = time.time()

            if self.config.as_enabled and songinfo:
                # No need to check if the song is 30 seconds or longer,
                # audioscrobbler.py takes care of that.
                if 'time' in songinfo:
                    self.np(songinfo)

                    self.scrob_start_time = str(int(time.time()))
                    self.scrob_last_prepared = songinfo.file

    def np(self, songinfo):
        thread = threading.Thread(target=self.do_np, args=(songinfo,))
        thread.name = "ScrobblerNowPlaying"
        thread.daemon = True
        thread.start()

    def do_np(self, songinfo):
        self.init()
        if self.config.as_enabled and self.scrob_post and songinfo:
            if 'artist' in songinfo and \
               'title' in songinfo and \
               'time' in songinfo:
                album = songinfo.get('album', '')
                tracknumber = songinfo.get('track', '')
                try:
                    self.scrob_post.nowplaying(songinfo.artist,
                                               songinfo.title,
                                               songinfo.time,
                                               tracknumber,
                                               album)
                except:
                    self.logger.exception(
                        "Unable to send 'now playing' data to the scrobbler")
        time.sleep(10)

    def post(self, prevsonginfo):
        self.init()
        if self.config.as_enabled and self.scrob_post and prevsonginfo:
            if 'artist' in prevsonginfo and \
               'title' in prevsonginfo and \
               'time' in prevsonginfo:
                album = prevsonginfo.get('album', '')
                tracknumber = prevsonginfo.get('track', '')
                try:
                    self.scrob_post.addtrack(
                        prevsonginfo.artist,
                        prevsonginfo.title,
                        prevsonginfo.time,
                        self.scrob_start_time,
                        tracknumber,
                        album)
                except:
                    self.logger.critical("Unable to add track to scrobbler")

                thread = threading.Thread(target=self.do_post)
                thread.name = "ScrobblerPost"
                thread.daemon = True
                thread.start()
        self.scrob_start_time = ""

    def do_post(self):
        for _i in range(0, 3):
            if not self.scrob_post:
                return
            if len(self.scrob_post.cache) == 0:
                return
            try:
                self.scrob_post.post()
            except audioscrobbler.AudioScrobblerConnectionError as e:
                self.logger.exception(
                    "Error while posting data to the scrobbler")
            time.sleep(10)

    def save_cache(self):
        """Save the cache in a file"""
        filename = os.path.expanduser('~/.config/sonata/ascache')
        if self.scrob_post:
            self.scrob_post.savecache(filename)

    def retrieve_cache(self):
        filename = os.path.expanduser('~/.config/sonata/ascache')
        if self.scrob_post:
            self.scrob_post.retrievecache(filename)




# this is the magic interpreted by Sonata, referring to on_enable etc. below:

### BEGIN PLUGIN INFO
# [plugin]
# plugin_format: 0, 0
# name: Test plugin
# version: 0, 0, 1
# description: A simple test plugin.
# author: Tuukka Hastrup
# author_email: Tuukka.Hastrup@iki.fi
# url: http://www.nongnu.org/sonata/
# license: GPL v3 or later
# [capabilities]
# enablables: on_enable
# tab_construct: tab_construct
# playing_song_observers: on_song_change
# lyrics_fetching: on_lyrics_fetch
### END PLUGIN INFO

# nothing magical from here on

from gi.repository import Gtk, Gdk, GdkPixbuf, GObject, Pango

from sonata.misc import escape_html

songlabel = None
lyricslabel = None

# this gets called when the plugin is loaded, enabled, or disabled:
def on_enable(state):
    global songlabel, lyricslabel
    if state:
        songlabel = Gtk.Label("No song info received yet.")
        songlabel.props.ellipsize = Pango.ELLIPSIZE_END
        lyricslabel = Gtk.Label("No lyrics requests yet.")
        lyricslabel.props.ellipsize = Pango.ELLIPSIZE_END
    else:
        songlabel = None
        lyricslabel = None

# this constructs the parts of the tab when called:
def tab_construct():
    vbox = Gtk.VBox()
    vbox.pack_start(Gtk.Label("Hello world!"), True, True, 0)
    vbox.pack_start(songlabel, True, True, 0)
    vbox.pack_start(lyricslabel, True, True, 0)
    vbox.pack_start(Gtk.Label("(You can modify me at %s)" %
                  __file__.rstrip("c")), True, True, 0)
    vbox.show_all()

    # the return value goes off to Base.add_tab(page, stock, text, focus):
    # (tab content, icon name, tab name, the widget to focus on tab switch)
    return (vbox, None, "Test plugin", None)

# this gets called when a new song is playing:
def on_song_change(songinfo):
    if songinfo:
        songlabel.set_markup("<b>Info for currently playing song:</b>"+
                     "\n%s" % escape_html(repr(songinfo)))
    else:
        songlabel.set_text("Currently not playing any song.")
    songlabel.show()

# this gets requests for lyrics:
def on_lyrics_fetch(callback, artist, title):
    lyricslabel.set_markup(
        "Got request for lyrics for artist %r title %r." %
        (artist, title))

    # callback(lyrics, error)
    GLib.timeout_add(
        0, callback, None,
        "%s doesn't have lyrics for %r." % (__name__, (artist, title)))



### BEGIN PLUGIN INFO
# [plugin]
# name: Last.fm covers
# plugin_format: 0, 0
# version: 1, 0, 0
# description: Fetch album covers from www.last.fm
# author: Jonathan Ballet
# author_email: jon@multani.info
# url: https://github.com/multani/sonata
# license: GPL v3 or later
# [capabilities]
# cover_fetching: on_cover_fetch
### END PLUGIN INFO

import json
import logging
import shutil
import urllib
from urllib import error, request, parse

from sonata.version import version


API_KEY = "41a1b04ed273fe997d2fddc3823dfb0f"
logger = logging.getLogger(__name__)


def make_user_agent():
    return "Sonata/%s +https://github.com/multani/sonata/" % version


def on_cover_fetch(artist, album, on_save_cb, on_err_cb):
    handler = urllib.request.HTTPHandler()
    opener = urllib.request.build_opener(handler)
    opener.addheaders = [("User-Agent", make_user_agent())]

    # First, find the link to the master release of this album
    search_url = "http://ws.audioscrobbler.com/2.0/?%s" % (
        urllib.parse.urlencode({
            "method": "album.getInfo",
            "artist": artist,
            "album": album,
            "api_key": API_KEY,
            "format": "json",
        }))

    logger.debug("Querying %r...", search_url)
    response = opener.open(search_url)
    lastfm = json.loads(response.read().decode('utf-8'))

    if 'error' in lastfm:
        logger.warning("Can't find cover on Last.fm: %s (err=%d)",
                       lastfm['message'], lastfm['error'])
        return

    for image in lastfm['album']['image']:
        if image['size'] != 'mega':
            continue

        url = image['#text']
        if url == '':
            logger.info("Found an album on Last.fm, but no cover :( %s",
                        lastfm['album']['url'])
            continue

        logger.debug("Downloading %r", url)
        try:
            response = opener.open(url)
        except urllib.error.URLError as e:
            logger.warning("Can't download %r: %s", url, e)
            if on_err_cb("Can't download %r: %s" % (url, e)):
                break
            else:
                continue

        if not on_save_cb(response):
            break


if __name__ == '__main__':
    import os
    import subprocess
    import tempfile
    logging.basicConfig(level=logging.DEBUG)
    fp, dest = tempfile.mkstemp(".png")

    def on_save(data):
        os.write(fp, data.read())
        os.close(fp)
        return False

    def on_err():
        print("Error!")
        return True

    result = on_cover_fetch("Metallica", "Ride the lightning", on_save, on_err)
    print(dest)
    subprocess.call(['xdg-open', dest])



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""
This plugin implements D-Bus features for Sonata:

 * Check that only one instance of Sonata is running at a time
 * Allow other programs to request the info popup, and to show or to toggle
   the main window visibility
 * Listen to Gnome 2.18+ multimedia key events

XXX Not a real plugin yet.

Example usage:
import dbus_plugin as dbus
self.dbus_service = dbus.SonataDBus(self.dbus_show, self.dbus_toggle,
                                    self.dbus_popup, self.dbus_fullscreen)
dbus.init_gnome_mediakeys(self.mpd_pp, self.mpd_stop, self.mpd_prev,
                            self.mpd_next)
if not dbus.using_gnome_mediakeys():
        # do something else instead...
"""

import logging
import sys

try:
    import dbus
    import dbus.service
    import _dbus_bindings as dbus_bindings
    from dbus.mainloop.glib import DBusGMainLoop
    DBusGMainLoop(set_as_default=True)
    HAVE_DBUS = True
except:
    HAVE_DBUS = False


logger = logging.getLogger(__name__)


def using_dbus():
    return HAVE_DBUS

HAVE_GNOME_MMKEYS = False


def using_gnome_mediakeys():
    return HAVE_GNOME_MMKEYS


def init_gnome_mediakeys(mpd_pp, mpd_stop, mpd_prev, mpd_next):
    global HAVE_GNOME_MMKEYS
    if HAVE_DBUS:
        try:
            bus = dbus.SessionBus()
            dbusObj = bus.get_object('org.freedesktop.DBus',
                                     '/org/freedesktop/DBus')
            dbusInterface = dbus.Interface(dbusObj,
                                            'org.freedesktop.DBus')
            if dbusInterface.NameHasOwner('org.gnome.SettingsDaemon'):
                try:
                    # mmkeys for gnome 2.22+
                    settingsDaemonObj = bus.get_object(
                        'org.gnome.SettingsDaemon',
                        '/org/gnome/SettingsDaemon/MediaKeys')
                    settingsDaemonInterface = dbus.Interface(settingsDaemonObj,
                                        'org.gnome.SettingsDaemon.MediaKeys')
                    settingsDaemonInterface.GrabMediaPlayerKeys('Sonata', 0)
                except:
                    # mmkeys for gnome 2.18+
                    settingsDaemonObj = bus.get_object(
                        'org.gnome.SettingsDaemon',
                        '/org/gnome/SettingsDaemon')
                    settingsDaemonInterface = dbus.Interface(settingsDaemonObj,
                                                    'org.gnome.SettingsDaemon')
                    settingsDaemonInterface.GrabMediaPlayerKeys('Sonata', 0)
                settingsDaemonInterface.connect_to_signal(
                    'MediaPlayerKeyPressed', lambda app,
                    key: mediaPlayerKeysCallback(mpd_pp, mpd_stop, mpd_prev,
                                                mpd_next, app, key))
                HAVE_GNOME_MMKEYS = True
        except:
            pass


def mediaPlayerKeysCallback(mpd_pp, mpd_stop, mpd_prev, mpd_next, app, key):
    if app == 'Sonata':
        if key in ('Play', 'PlayPause', 'Pause'):
            mpd_pp(None)
        elif key == 'Stop':
            mpd_stop(None)
        elif key == 'Previous':
            mpd_prev(None)
        elif key == 'Next':
            mpd_next(None)


def get_session_bus():
    try:
        return dbus.SessionBus()
    except Exception:
        logger.error(
            _('Sonata failed to connect to the D-BUS session bus: '
              'Unable to determine the address of the message bus '
              '(try \'man dbus-launch\' and \'man dbus-daemon\' '
              'for help)")'))
        raise


def execute_remote_commands(toggle=False, popup=False, fullscreen=False,
                            start=False):
    try:
        bus = get_session_bus()
        obj = bus.get_object('org.MPD', '/org/MPD/Sonata')
        if toggle:
            obj.toggle(dbus_interface='org.MPD.SonataInterface')
        if popup:
            obj.popup(dbus_interface='org.MPD.SonataInterface')
        if fullscreen:
            obj.fullscreen(dbus_interface='org.MPD.SonataInterface')
        sys.exit()
    except Exception:
        logger.warning(_("Failed to execute remote commands."))
        if start is None or start:
            logger.info(_("Starting Sonata instead..."))
        else:
            # TODO: should we log the exception here?
            logger.critical(_("Maybe Sonata is not running?"))
            sys.exit(1)


if HAVE_DBUS:

    class SonataDBus(dbus.service.Object):

        def __init__(self, dbus_show, dbus_toggle, dbus_popup, dbus_fullscreen):
            self.dbus_show = dbus_show
            self.dbus_toggle = dbus_toggle
            self.dbus_popup = dbus_popup
            self.dbus_fullscreen = dbus_fullscreen
            session_bus = get_session_bus()
            bus_name = dbus.service.BusName('org.MPD', bus=session_bus)
            object_path = '/org/MPD/Sonata'
            super().__init__(bus_name, object_path)

        @dbus.service.method('org.MPD.SonataInterface')
        def show(self):
            self.dbus_show()

        @dbus.service.method('org.MPD.SonataInterface')
        def toggle(self):
            self.dbus_toggle()

        @dbus.service.method('org.MPD.SonataInterface')
        def popup(self):
            self.dbus_popup()

        @dbus.service.method('org.MPD.SonataInterface')
        def fullscreen(self):
            self.dbus_fullscreen()



# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides a user interface for editing the metadata tags of
local music files.

Example usage:
import tagedit
tageditor = tagedit.TagEditor(self.window, self.tags_mpd_update)
tageditor.on_tags_edit(files, temp_mpdpaths, self.musicdir[self.profile_num])
"""

import os
import re

from gi.repository import Gtk, GLib
tagpy = None # module loaded when needed

from sonata import ui, misc


class TagEditor:
    """This class implements a dialog for editing music metadata tags.

    When the dialog closes, the callback gets the list of updates made.
    """
    def __init__(self, window, tags_mpd_update, tags_set_use_mpdpath):
        self.window = window
        self.tags_mpd_update = tags_mpd_update
        self.tags_set_use_mpdpath = tags_set_use_mpdpath

        self.file_label = None
        self.curr_mpdpath = None
        self.tagnum = None
        self.use_mpdpaths = None

        self.edit_window = None
        self.entries = None
        self.tags = None

        self.builder = ui.builder('tagedit')
        self.css_provider = ui.css_provider('tagedit')


    def _init_edit_window(self):
        self.edit_window = self.builder.get_object('tags_dialog')
        self.edit_window.set_transient_for(self.window)
        self.file_label = self.builder.get_object('tags_file_label')
        genre_combo = self.builder.get_object('tags_genre_comboboxtext')
        for genre in self.tags_win_genres():
            genre_combo.append_text(genre)
        expand_arrow = self.builder.get_object('tags_expand_arrow')
        expand_button = self.builder.get_object('tags_expand_file_button')
        self.set_expandbutton_state(expand_button, expand_arrow)
        expand_button.connect('clicked', self.toggle_path, expand_arrow)

        year_entry = self.builder.get_object('tags_year_entry')
        year_entry.connect('insert_text', self.tags_win_entry_constraint, True)
        track_entry = self.builder.get_object('tags_track_entry')
        track_entry.connect('insert_text', self.tags_win_entry_constraint, False)
        artist_entry = self.builder.get_object('tags_artist_entry')
        album_entry = self.builder.get_object('tags_album_entry')
        title_entry = self.builder.get_object('tags_title_entry')
        genre_entry = self.builder.get_object('tags_genre_entry')
        comment_entry = self.builder.get_object('tags_comment_entry')
        self.entries = {
            'title': title_entry,
            'artist': artist_entry,
            'album': album_entry,
            'year': year_entry,
            'track': track_entry,
            'genre': genre_entry,
            'comment': comment_entry,}

        year_button = self.builder.get_object('tags_save_year_button')
        track_button = self.builder.get_object('tags_save_track_button')
        artist_button = self.builder.get_object('tags_save_artist_button')
        album_button = self.builder.get_object('tags_save_album_button')
        title_button = self.builder.get_object('tags_save_title_button')
        genre_button = self.builder.get_object('tags_save_genre_button')
        comment_button = self.builder.get_object('tags_save_comment_button')
        buttons = (title_button, artist_button, album_button, year_button,
                   track_button, genre_button, comment_button)
        names = ('title', 'artist', 'album', 'year',
                 'track', 'genre', 'comment')
        for name, button in zip(names, buttons):
            entry = self.entries[name]
            entry.connect('changed', self.tags_win_entry_changed)
            button.connect('clicked', self.tags_win_apply_all, name, entry)

    def on_tags_edit(self, files, temp_mpdpaths, music_dir):
        """Display the editing dialog"""
        # Try loading module
        global tagpy
        if tagpy is None:
            try:
                import tagpy
            except ImportError:
                ui.show_msg(self.window, _("Taglib and/or tagpy not found, tag editing support disabled."), _("Edit Tags"), 'editTagsError', Gtk.ButtonsType.CLOSE, response_cb=ui.dialog_destroy)
                ui.change_cursor(None)
                return
            # Set default tag encoding to utf8.. fixes some reported bugs.
            import tagpy.id3v2 as id3v2
            id3v2.FrameFactory.instance().setDefaultTextEncoding(tagpy.StringType.UTF8)

        # Make sure tagpy is at least 0.91
        if hasattr(tagpy.Tag.title, '__call__'):
            ui.show_msg(self.window, _("Tagpy version < 0.91. Please upgrade to a newer version, tag editing support disabled."), _("Edit Tags"), 'editTagsError', Gtk.ButtonsType.CLOSE, response_cb=ui.dialog_destroy)
            ui.change_cursor(None)
            return

        if not os.path.isdir(music_dir):
            ui.show_msg(self.window, _("The path %s does not exist. Please specify a valid music directory in preferences.") % music_dir, _("Edit Tags"), 'editTagsError', Gtk.ButtonsType.CLOSE, response_cb=ui.dialog_destroy)
            ui.change_cursor(None)
            return

                # XXX file list was created here

        if len(files) == 0:
            ui.change_cursor(None)
            return

        # Initialize:
        self.tagnum = -1

        self.tags = [{
             'title': '', 'artist': '', 'album': '', 'year': '', 'track': '',
             'genre': '', 'comment': '', 'title-changed': False,
             'artist-changed': False, 'album-changed': False,
             'year-changed': False, 'track-changed': False,
             'genre-changed': False, 'comment-changed': False,
             'fullpath': filename,
             'mpdpath': path,}
            for filename, path in zip(files, temp_mpdpaths)]

        if not os.path.exists(self.tags[0]['fullpath']):
            ui.change_cursor(None)
            ui.show_msg(self.window, _("File '%s' not found. Please specify a valid music directory in preferences.") % self.tags[0]['fullpath'], _("Edit Tags"), 'editTagsError', Gtk.ButtonsType.CLOSE, response_cb=ui.dialog_destroy)
            return
        if not self.tags_next_tag():
            ui.change_cursor(None)
            ui.show_msg(self.window, _("No music files with editable tags found."), _("Edit Tags"), 'editTagsError', Gtk.ButtonsType.CLOSE, response_cb=ui.dialog_destroy)
            return
        if not self.edit_window:
            self._init_edit_window()

        saveall_button = self.builder.get_object('tags_saveall_button')
        if len(files) > 1:
            # Only show save all button if more than one song being edited.
            saveall_button.show()
            saveall_button.set_property("visible", True)
        else:
            saveall_button.hide()
            saveall_button.set_property("visible", False)

        self.tags_win_update()
        self.edit_window.show_all()
        SAVE_ALL = -12
        done = False

        while not done:
            # Next file:
            self.tags_win_update()
            response = self.edit_window.run()
            if response == SAVE_ALL:
                self.save_tag()
                while self.tags_next_tag():
                    self.tags_win_update()
                    self.save_tag()
                done = True
            elif response == Gtk.ResponseType.ACCEPT:
                self.save_tag()
                done = not self.tags_next_tag()
                if done:
                    # To ensure we update the last file in tags_mpd_update
                    self.tagnum = self.tagnum + 1
            elif response == Gtk.ResponseType.REJECT:
                done = True

        tag_paths = (tag['mpdpath'] for tag in self.tags[:self.tagnum])
        GLib.idle_add(self.tags_mpd_update, tag_paths)
        self.tags_set_use_mpdpath(self.use_mpdpaths)

        self.tags = None
        ui.change_cursor(None)
        self.edit_window.hide()

    def tags_next_tag(self):
        # Returns true if next tag found (and self.tagnum is updated).
        # If no next tag found, returns False.
        while self.tagnum < len(self.tags) - 1:
            self.tagnum = self.tagnum + 1
            if os.path.exists(self.tags[self.tagnum]['fullpath']):
                fileref = tagpy.FileRef(self.tags[self.tagnum]['fullpath'])
                if not fileref.isNull():
                    return True
        return False

    def tags_win_entry_changed(self, entry):
        style_context = entry.get_style_context()
        style_context.add_class('modified')

    def tags_win_entry_revert_color(self, entry):
        style_context = entry.get_style_context()
        style_context.remove_class('modified')

    def tags_win_apply_all(self, _button, item, entry):
        for tagnum, tag in enumerate(self.tags):
            tagnum = tagnum + 1
            if item in ("title", "album", "artist", "genre", "comment"):
                tag[item] = entry.get_text()
                tag[item + '-changed'] = True
            elif item == "year":
                if len(entry.get_text()) > 0:
                    tag['year'] = int(entry.get_text())
                else:
                    tag['year'] = 0
                tag['year-changed'] = True
            elif item == "track":
                if tagnum >= self.tagnum - 1:
                    # Start the current song at track 1, as opposed to the first
                    # song in the list.
                    tag['track'] = tagnum - self.tagnum
                tag['track-changed'] = True
        if item == "track":
            # Update the entry for the current song:
            entry.set_text(str(self.tags[self.tagnum]['track']))

    def tags_win_update(self):
        current_tag = self.tags[self.tagnum]
        tag = tagpy.FileRef(current_tag['fullpath']).tag()
        # Update interface:
        for entry_name, entry in self.entries.items():
            # Only retrieve info from the file if the info hasn't changed
            if not current_tag[entry_name + "-changed"]:
                current_tag[entry_name] = getattr(tag, entry_name, '')
            tag_value = current_tag[entry_name]
            if tag_value == 0:
                tag_value = ''
            try:
                entry.set_text(str(tag_value).strip())
            except AttributeError:
                pass

            # Revert text color if this tag wasn't changed by the user
            if not current_tag[entry_name + "-changed"]:
                self.tags_win_entry_revert_color(entry)

        self.curr_mpdpath = GLib.filename_display_name(current_tag['mpdpath'])
        filename = self.curr_mpdpath
        if not self.use_mpdpaths:
            filename = os.path.basename(filename)
        self.file_label.set_text(filename)
        self.entries['title'].grab_focus()
        self.edit_window.set_title(_("Edit Tags - %s of %s") %
            (self.tagnum + 1, len(self.tags)))

    def save_tag(self):
        filetag = tagpy.FileRef(self.tags[self.tagnum]['fullpath'])
        tag = filetag.tag()
        # Set tag fields according to entry text
        for field, entry in self.entries.items():
            tag_value = entry.get_text().strip()
            if field in ('year', 'track'):
                if len(tag_value) == 0:
                    tag_value = '0'
                tag_value = int(tag_value)
            if field == 'comment':
                if len(tag_value) == 0:
                    tag_value = ' '
            setattr(tag, field, tag_value)

        save_success = filetag.save()
        if not (save_success): # FIXME: was (save_success and self.conn and self.status):
            ui.show_msg(self.window,
                        _("Unable to save tag to music file."),
                        _("Edit Tags"), 'editTagsError',
                        Gtk.ButtonsType.CLOSE,
                        response_cb=ui.dialog_destroy)

    def tags_win_genres(self):
        return ["", "A Cappella", "Acid", "Acid Jazz", "Acid Punk", "Acoustic",
                "Alt. Rock", "Alternative", "Ambient", "Anime", "Avantgarde", "Ballad",
                "Bass", "Beat", "Bebob", "Big Band", "Black Metal", "Bluegrass",
                "Blues", "Booty Bass", "BritPop", "Cabaret", "Celtic", "Chamber music",
                "Chanson", "Chorus", "Christian Gangsta Rap", "Christian Rap",
                "Christian Rock", "Classic Rock", "Classical", "Club", "Club-House",
                "Comedy", "Contemporary Christian", "Country", "Crossover", "Cult",
                "Dance", "Dance Hall", "Darkwave", "Death Metal", "Disco", "Dream",
                "Drum & Bass", "Drum Solo", "Duet", "Easy Listening", "Electronic",
                "Ethnic", "Euro-House", "Euro-Techno", "Eurodance", "Fast Fusion",
                "Folk", "Folk-Rock", "Folklore", "Freestyle", "Funk", "Fusion", "Game",
                "Gangsta", "Goa", "Gospel", "Gothic", "Gothic Rock", "Grunge",
                "Hard Rock", "Hardcore", "Heavy Metal", "Hip-Hop", "House", "Humour",
                "Indie", "Industrial", "Instrumental", "Instrumental pop",
                "Instrumental rock", "JPop", "Jazz", "Jazz+Funk", "Jungle", "Latin",
                "Lo-Fi", "Meditative", "Merengue", "Metal", "Musical", "National Folk",
                "Native American", "Negerpunk", "New Age", "New Wave", "Noise",
                "Oldies", "Opera", "Other", "Polka", "Polsk Punk", "Pop", "Pop-Folk",
                "Pop/Funk", "Porn Groove", "Power Ballad", "Pranks", "Primus",
                "Progressive Rock", "Psychedelic", "Psychedelic Rock", "Punk",
                "Punk Rock", "R&B", "Rap", "Rave", "Reggae", "Retro", "Revival",
                "Rhythmic soul", "Rock", "Rock & Roll", "Salsa", "Samba", "Satire",
                "Showtunes", "Ska", "Slow Jam", "Slow Rock", "Sonata", "Soul",
                "Sound Clip", "Soundtrack", "Southern Rock", "Space", "Speech",
                "Swing", "Symphonic Rock", "Symphony", "Synthpop", "Tango", "Techno",
                "Techno-Industrial", "Terror", "Thrash Metal", "Top 40", "Trailer"]

    def tags_win_entry_constraint(self, entry, new_text, _new_text_length, _broken_position, isyearlabel):
        entry_chars = entry.get_text()
        pos = entry.get_position()
        proposed_text = entry_chars[:pos] + new_text + entry_chars[pos:]

        # get the correct regular expression
        expr = r'(0|[1-9][0-9]{0,3})$' if isyearlabel else r'(0|[1-9][0-9]*)$'
        expr = re.compile(expr)

        if not expr.match(proposed_text):
            # deny
            entry.stop_emission("insert-text")

    def toggle_path(self, button, arrow):
        self.use_mpdpaths = not self.use_mpdpaths
        if self.use_mpdpaths:
            self.file_label.set_text(self.curr_mpdpath)
        else:
            self.file_label.set_text(os.path.basename(self.curr_mpdpath))
        self.set_expandbutton_state(button, arrow)

    def set_expandbutton_state(self, button, arrow):
        if self.use_mpdpaths:
            arrow.set_property("arrow-type", Gtk.ArrowType.LEFT)
            button.set_tooltip_text(_("Hide file path"))
        else:
            arrow.set_property("arrow-type", Gtk.ArrowType.RIGHT)
            button.set_tooltip_text(_("Show file path"))

    def set_use_mpdpaths(self, use_mpdpaths):
        self.use_mpdpaths = use_mpdpaths




# Copyright 2006-2009 Scott Horowitz <stonecrest@gmail.com>
# Copyright 2009-2014 Jonathan Ballet <jon@multani.info>
#
# This file is part of Sonata.
#
# Sonata is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Sonata is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Sonata.  If not, see <http://www.gnu.org/licenses/>.

import collections
import logging
import os
import re
import shutil
import threading # artwork_update starts a thread _artwork_update

from gi.repository import Gtk, Gdk, GdkPixbuf, GLib, GObject

from sonata import img, ui, misc, consts, mpdhelper as mpdh
from sonata.song import SongRecord
from sonata.pluginsystem import pluginsystem


COVERS_DIR = os.path.expanduser("~/.covers")
COVERS_TEMP_DIR = os.path.join(COVERS_DIR, 'temp')
logger = logging.getLogger(__name__)


class ArtworkLocator:
    """Find and return artwork paths requested for songs."""

    def __init__(self, config):
        self.config = config

    def _get_locations(self, artist, album, song_dir, default_kind=None):
        """Get the various possible locations for the artwork for one song."""

        artist = (artist or "").replace("/", "")
        album = (album or "").replace("/", "")
        song_dir = os.path.join(
            self.config.current_musicdir,
            get_multicd_album_root_dir(song_dir)
        )
        join = os.path.join
        custom_filename = self.config.art_location_custom_filename

        # Build a map of location kinds and actual locations. The order is
        # important and represents the most common/easiest to the least
        # common/more expensive ways of finding covers.
        # Each location is an iterable of possibly several locations.
        pre_map = [
            ('HOMECOVERS', (COVERS_DIR, "%s-%s.jpg" % (artist, album))),
            ('COVER',      (song_dir, "cover.jpg")),
            ('FOLDER',     (song_dir, "folder.jpg")),
            ('ALBUM',      (song_dir, "album.jpg")),
            ('CUSTOM',     ((song_dir, custom_filename)
                            if len(custom_filename) > 0
                            else []
                           )),
            ('MISC',       (join(song_dir, location)
                            for location in consts.ART_LOCATIONS_MISC)),
            ('SINGLE',     self._lookup_single_image(song_dir)),
        ]

        map = collections.OrderedDict()
        for (fake_key, value) in pre_map:
            key = getattr(consts, 'ART_LOCATION_%s' % fake_key)
            if isinstance(value, tuple):
                value = [join(*value)]
            map[key] = value

        # Move the default kind requested to the beginning of the map, so it has
        # priority over the other.
        if default_kind in map:
            map.move_to_end(default_kind, last=False)

        return map

    def _lookup_single_image(self, song_dir):
        """Look up a song directory to find one image to be the artwork.

        This basically loops over all the files in a song directory and returns
        something only if it found exactly one image. This will be the artwork
        image.
        """

        try:
            files = os.listdir(song_dir)
        except OSError:
            return

        get_ext = lambda path: os.path.splitext(path)[1][1:]

        artworks = [f for f in files if get_ext(f) in img.VALID_EXTENSIONS]
        if len(artworks) != 1:
            return

        yield os.path.join(song_dir, artworks[0])

    def path(self, artist, album, song_dir, specific_kind=None):
        """Return the artwork path from a song data (without checks)

        By default, the location of the song is made based on the preferences
        set by the user, but this can be overriden if necessary.
        No guarantees are made with the respect to the path validity.
        """

        locations_map = self._get_locations(artist, album, song_dir,
                                            self.config.art_location)

        if specific_kind is None:
            specific_kind = self.config.art_location

        try:
            result = next(iter(locations_map[specific_kind]))
        except StopIteration: # We tried an empty location :(
            result = None
        return result


    def path_from_song(self, song, specific_kind=None):
        """Same as `path()` but using a Song object."""
        return self.path(song.artist, song.album,
                         os.path.dirname(song.file), specific_kind)

    def locate(self, artist, album, song_dir):
        """Locate an actual artwork for the specified data.

        Sonata tries *very* hard to find an artwork for the specified data. It
        looks in "well-known" places and return the first artwork which actually
        exists.
        """

        # XXX: the 'kind' returned can probably be removed once the calling code
        # will be refactored. It is used only for some kind of thread
        # concurrency management, and is probably buggy and not optimal...

        locations_map = self._get_locations(artist, album, song_dir,
                                            self.config.art_location)

        for kind, locations in locations_map.items():
            for location in locations:
                if os.path.exists(location):
                    return (kind, location)

        return (None, None)


def get_multicd_album_root_dir(albumpath):
    """Go one dir upper for multicd albums

    >>> from sonata.artwork import get_multicd_album_root_dir as f
    >>> f('Moonspell/1995 - Wolfheart/cd 2')
    'Moonspell/1995 - Wolfheart'
    >>> f('2007 - Dark Passion Play/CD3')
    '2007 - Dark Passion Play'
    >>> f('Ayreon/2008 - 01011001/CD 1 - Y')
    'Ayreon/2008 - 01011001'

    """

    if re.compile(r'(?i)cd\s*\d+').match(os.path.split(albumpath)[1]):
        albumpath = os.path.split(albumpath)[0]
    return albumpath


def artwork_path(song, config):
    if song.name is not None:
        f = artwork_stream(song.name)
    else:
        f = ArtworkLocator(config).path_from_song(song)
    return f


def artwork_stream(stream_name):
    return os.path.join(os.path.expanduser('~/.covers'), "%s.jpg" %
                        stream_name.replace("/", ""))


class Artwork(GObject.GObject):

    __gsignals__ = {
        'artwork-changed': (GObject.SIGNAL_RUN_FIRST, None,
                            (GdkPixbuf.Pixbuf,)),
        'artwork-reset': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self, config, is_lang_rtl, schedule_gc_collect,
                 imagelist_append, remotefilelist_append,
                 allow_art_search, status_is_play_or_pause,
                 album_image, tray_image):
        super().__init__()

        self.config = config
        self.locator = ArtworkLocator(config)
        self.album_filename = 'sonata-album'

        # constants from main
        self.is_lang_rtl = is_lang_rtl

        # callbacks to main XXX refactor to clear this list
        self.schedule_gc_collect = schedule_gc_collect
        self.imagelist_append = imagelist_append
        self.remotefilelist_append = remotefilelist_append
        self.allow_art_search = allow_art_search
        self.status_is_play_or_pause = status_is_play_or_pause

        # local pixbufs, image file names
        self.sonatacd = Gtk.IconFactory.lookup_default('sonata-cd')
        self.sonatacd_large = Gtk.IconFactory.lookup_default('sonata-cd-large')
        self.albumpb = None
        self.currentpb = None

        # local UI widgets provided to main by getter methods
        self.albumimage = album_image
        self.albumimage.set_from_icon_set(self.sonatacd, -1)

        self.tray_album_image = tray_image

        # local version of Main.songinfo mirrored by update_songinfo
        self.songinfo = None

        # local state
        self.lastalbumart = None
        self.single_img_in_dir = None
        self.misc_img_in_dir = None
        self.stop_art_update = False
        self.downloading_image = False
        self.lib_art_cond = None

        # local artwork, cache for library
        self.lib_model = None
        self.lib_art_rows_local = []
        self.lib_art_rows_remote = []
        self.lib_art_pb_size = 0

        self.cache = ArtworkCache(self.config)
        self.cache.load()

    def update_songinfo(self, songinfo):
        self.songinfo = songinfo

    def on_reset_image(self, _action):
        if self.songinfo:
            if 'name' in self.songinfo:
                # Stream, remove file:
                misc.remove_file(artwork_stream(self.songinfo.name))
            else:
                # Normal song:
                misc.remove_file(self.locator.path_from_song(self.songinfo))
                misc.remove_file(self.locator.path_from_song(
                    self.songinfo, consts.ART_LOCATION_HOMECOVERS))
                # Use blank cover as the artwork
                dest_filename = self.locator.path_from_song(
                    self.songinfo, consts.ART_LOCATION_HOMECOVERS)
                try:
                    emptyfile = open(dest_filename, 'w')
                    emptyfile.close()
                except IOError:
                    pass
            self.artwork_update(True)

    def artwork_set_tooltip_art(self, pix):
        # Set artwork
        pix = pix.new_subpixbuf(0, 0, 77, 77)
        self.tray_album_image.set_from_pixbuf(pix)
        del pix

    def artwork_stop_update(self):
        self.stop_art_update = True

    def artwork_is_downloading_image(self):
        return self.downloading_image

    def library_artwork_init(self, model, pb_size):

        self.lib_model = model
        self.lib_art_pb_size = pb_size

        self.lib_art_cond = threading.Condition()
        thread = threading.Thread(target=self._library_artwork_update)
        thread.name = "ArtworkLibraryUpdate"
        thread.daemon = True
        thread.start()

    def library_artwork_update(self, model, start_row, end_row, albumpb):
        self.albumpb = albumpb

        # Update self.lib_art_rows_local with new rows followed
        # by the rest of the rows.
        self.lib_art_cond.acquire()
        self.lib_art_rows_local = []
        self.lib_art_rows_remote = []
        start = start_row.get_indices()[0]
        end = end_row.get_indices()[0]
        test_rows = list(range(start, end + 1)) + list(range(len(model)))
        for row in test_rows:
            i = model.get_iter((row,))
            icon = model.get_value(i, 0)
            if icon == self.albumpb:
                data = model.get_value(i, 1)
                self.lib_art_rows_local.append((i, data, icon))
        self.lib_art_cond.notifyAll()
        self.lib_art_cond.release()

    def _library_artwork_update(self):
        while True:
            # Wait for items..
            self.lib_art_cond.acquire()
            while(len(self.lib_art_rows_local) == 0 and \
                  len(self.lib_art_rows_remote) == 0):
                self.lib_art_cond.wait()
            self.lib_art_cond.release()

            # Try first element, giving precedence to local queue:
            if self.lib_art_rows_local:
                i, data, icon = self.lib_art_rows_local.pop(0)
                cb = self._library_artwork_update_local
            elif self.lib_art_rows_remote:
                i, data, icon = self.lib_art_rows_remote.pop(0)
                cb = self._library_artwork_update_remote
            else:
                continue

            cache_key = SongRecord(artist=data.artist, album=data.album,
                                         path=data.path)
            # Try to replace default icons with cover art:
            pb = self.cache.get_pixbuf(cache_key, self.lib_art_pb_size)
            cb(i, data, icon, cache_key, pb)


    def _library_artwork_update_local(self, i, data, icon, cache_key, pb):
        filename = None

        if pb is not None:
            # Continue to rescan for local artwork if we are
            # displaying the default album image, in case the user
            # has added a local image since we first scanned.
            filename = self.cache.get(cache_key)
            if os.path.basename(filename) == os.path.basename(
                self.album_filename):
                filename = None
                pb = None

        # No cached pixbuf, try local/remote search:
        if pb is None:
            pb, filename = self.library_get_album_cover(
                data.path, data.artist, data.album, self.lib_art_pb_size)

        # Set pixbuf icon in model; add to cache
        if pb is not None and filename is not None:
            self.cache.set(cache_key, filename)
            GLib.idle_add(self.library_set_cover, i, pb, data)

        if pb is None and self.config.covers_pref == consts.ART_LOCAL_REMOTE:
            # No local art found, add to remote queue for later
            self.lib_art_rows_remote.append((i, data, icon))


    def _library_artwork_update_remote(self, i, data, icon, cache_key, pb):
        filename = None

        # No cached pixbuf, try local/remote search:
        if pb is None:
            pb, filename = self.library_get_album_cover(data.path,
                                                        data.artist,
                                                        data.album,
                                                        self.lib_art_pb_size)
            filename = self.locator.path(data.artist, data.album, data.path)
            self.artwork_download_img_to_file(data.artist, data.album,
                                              filename)

        # Set pixbuf icon in model; add to cache
        if pb is not None and filename is not None:
            self.cache.set(cache_key, filename)
            GLib.idle_add(self.library_set_cover, i, pb, data)

        if pb is None:
            # No remote art found, store self.albumpb filename in cache
            self.cache.set(cache_key, self.album_filename)


    def library_set_image_for_current_song(self, cache_key):
        # Search through the rows in the library to see
        # if we match the currently playing song:
        if cache_key.artist is None and cache_key.album is None:
            return
        for row in self.lib_model:
            if str(cache_key.artist).lower() == str(row[1].artist).lower() \
            and str(cache_key.album).lower() == str(row[1].album).lower():
                pb = self.cache.get_pixbuf(cache_key, self.lib_art_pb_size)
                if pb:
                    self.lib_model.set_value(row.iter, 0, pb)

    def library_set_cover(self, i, pb, data):
        if self.lib_model.iter_is_valid(i):
            if self.lib_model.get_value(i, 1) == data:
                self.lib_model.set_value(i, 0, pb)

    def library_get_album_cover(self, song_dir, artist, album, pb_size):
        _tmp, coverfile = self.locator.locate(artist, album, song_dir)
        if coverfile:
            try:
                coverpb = GdkPixbuf.Pixbuf.new_from_file_at_size(coverfile,
                                                            pb_size, pb_size)
            except:
                # Delete bad image:
                misc.remove_file(coverfile)
                return (None, None)
            w = coverpb.get_width()
            h = coverpb.get_height()
            coverpb = img.do_style_cover(self.config, coverpb, w, h)
            return (coverpb, coverfile)
        return (None, None)

    def artwork_update(self, force=False):
        if force:
            self.lastalbumart = None

        self.stop_art_update = False
        if not self.config.show_covers:
            return
        if not self.songinfo:
            self.artwork_set_default_icon()
            return

        if self.status_is_play_or_pause():
            thread = threading.Thread(target=self._artwork_update)
            thread.name = "ArtworkUpdate"
            thread.daemon = True
            thread.start()
        else:
            self.artwork_set_default_icon()

    def _artwork_update(self):
        if 'name' in self.songinfo:
            # Stream
            streamfile = artwork_stream(self.songinfo.name)
            if os.path.exists(streamfile):
                GLib.idle_add(self.artwork_set_image, streamfile, None, None,
                              None)
            else:
                self.artwork_set_default_icon()
        else:
            # Normal song:
            artist = self.songinfo.artist or ""
            album = self.songinfo.album or ""
            path = os.path.dirname(self.songinfo.file)
            if len(artist) == 0 and len(album) == 0:
                self.artwork_set_default_icon(artist, album, path)
                return
            filename = self.locator.path_from_song(self.songinfo)
            if filename == self.lastalbumart:
                # No need to update..
                self.stop_art_update = False
                return
            self.lastalbumart = None
            imgfound = self.artwork_check_for_local(artist, album, path)
            if not imgfound:
                if self.config.covers_pref == consts.ART_LOCAL_REMOTE:
                    imgfound = self.artwork_check_for_remote(artist, album,
                                                             path, filename)

    def artwork_check_for_local(self, artist, album, path):
        self.artwork_set_default_icon(artist, album, path)
        self.misc_img_in_dir = None
        self.single_img_in_dir = None
        location_type, filename = self.locator.locate(
            self.songinfo.artist, self.songinfo.album,
            os.path.dirname(self.songinfo.file))

        if location_type is not None and filename:
            if location_type == consts.ART_LOCATION_MISC:
                self.misc_img_in_dir = filename
            elif location_type == consts.ART_LOCATION_SINGLE:
                self.single_img_in_dir = filename
            GLib.idle_add(self.artwork_set_image, filename, artist, album, path)
            return True

        return False

    def artwork_check_for_remote(self, artist, album, path, filename):
        self.artwork_set_default_icon(artist, album, path)
        self.artwork_download_img_to_file(artist, album, filename)
        if os.path.exists(filename):
            GLib.idle_add(self.artwork_set_image, filename, artist, album, path)
            return True
        return False

    def artwork_set_default_icon(self, artist=None, album=None, path=None):
        GLib.idle_add(self.albumimage.set_from_icon_set,
                      self.sonatacd, -1)
        self.emit('artwork-reset')
        GLib.idle_add(self.tray_album_image.set_from_icon_set,
                      self.sonatacd, -1)

        self.lastalbumart = None

        # Also, update row in library:
        if artist is not None:
            cache_key = SongRecord(artist=artist, album=album, path=path)
            self.cache.set(cache_key, self.album_filename)
            GLib.idle_add(self.library_set_image_for_current_song, cache_key)

    def artwork_set_image(self, filename, artist, album, path,
                          info_img_only=False):
        # Note: filename arrives here is in FILESYSTEM_CHARSET, not UTF-8!
        if self.artwork_is_for_playing_song(filename):
            if os.path.exists(filename):

                # We use try here because the file might exist, but might
                # still be downloading or corrupt:
                try:
                    pix = GdkPixbuf.Pixbuf.new_from_file(filename)
                except:
                    # If we have a 0-byte file, it should mean that
                    # sonata reset the image file. Otherwise, it's a
                    # bad file and should be removed.
                    if os.stat(filename).st_size != 0:
                        misc.remove_file(filename)
                    return

                self.currentpb = pix

                if not info_img_only:
                    # Store in cache
                    cache_key = SongRecord(artist=artist, album=album,
                                           path=path)
                    self.cache.set(cache_key, filename)

                    # Artwork for tooltip, left-top of player:
                    (pix1, w, h) = img.get_pixbuf_of_size(pix, 75)
                    pix1 = img.do_style_cover(self.config, pix1, w, h)
                    pix1 = img.pixbuf_add_border(pix1)
                    pix1 = img.pixbuf_pad(pix1, 77, 77)
                    self.albumimage.set_from_pixbuf(pix1)
                    self.artwork_set_tooltip_art(pix1)
                    del pix1

                    # Artwork for library, if current song matches:
                    self.library_set_image_for_current_song(cache_key)

                self.emit('artwork-changed', pix)
                del pix

                self.lastalbumart = filename

                self.schedule_gc_collect()

    def artwork_set_image_last(self):
        self.artwork_set_image(self.lastalbumart, None, None, None, True)

    def artwork_is_for_playing_song(self, filename):
        # Since there can be multiple threads that are getting album art,
        # this will ensure that only the artwork for the currently playing
        # song is displayed
        if self.status_is_play_or_pause() and self.songinfo:
            if 'name' in self.songinfo:
                streamfile = artwork_stream(self.songinfo.name)
                if filename == streamfile:
                    return True
            else:
                # Normal song:
                if (filename in [self.locator.path_from_song(self.songinfo, l)
                                 for l in [consts.ART_LOCATION_HOMECOVERS,
                                           consts.ART_LOCATION_COVER,
                                           consts.ART_LOCATION_ALBUM,
                                           consts.ART_LOCATION_FOLDER,
                                           consts.ART_LOCATION_CUSTOM]] or
                    (self.misc_img_in_dir and \
                     filename == self.misc_img_in_dir) or
                    (self.single_img_in_dir and filename == \
                     self.single_img_in_dir)):
                    return True
        # If we got this far, no match:
        return False

    def artwork_download_img_to_file(self, artist, album, dest_filename,
                                     all_images=False):

        downloader = CoverDownloader(dest_filename, self.download_progress,
                                     all_images)

        self.downloading_image = True
        # Fetch covers from covers websites or such...
        cover_fetchers = pluginsystem.get('cover_fetching')
        for plugin, callback in cover_fetchers:
            logger.info("Looking for covers for %r from %r (using %s)",
                        album, artist, plugin.name)

            try:
                callback(artist, album,
                         downloader.on_save_callback, downloader.on_err_cb)
            except Exception as e:
                if logger.isEnabledFor(logging.DEBUG):
                    log = logger.exception
                else:
                    log = logger.warning

                log("Error while downloading covers from %s: %s",
                    plugin.name, e)

            if downloader.found_images:
                break

        self.downloading_image = False
        return downloader.found_images

    def download_progress(self, dest_filename_curr, i):
        # This populates Main.imagelist for the remote image window
        if os.path.exists(dest_filename_curr):
            pix = GdkPixbuf.Pixbuf.new_from_file(dest_filename_curr)
            pix = pix.scale_simple(148, 148, GdkPixbuf.InterpType.HYPER)
            pix = img.do_style_cover(self.config, pix, 148, 148)
            pix = img.pixbuf_add_border(pix)
            if self.stop_art_update:
                del pix
                return False # don't continue to next image
            self.imagelist_append([i + 1, pix])
            del pix
            self.remotefilelist_append(dest_filename_curr)
            if i == 0:
                self.allow_art_search()

            ui.change_cursor(None)

        return True # continue to next image

    def have_last(self):
        if self.lastalbumart is not None:
            return True
        return False


class CoverDownloader:
    """Download covers and store them in temporary files"""

    def __init__(self, path, progress_cb, all_images):
        self.path = path
        self.progress_cb = progress_cb
        self.max_images = 50 if all_images else 1
        self.current = 0

    @property
    def found_images(self):
        return self.current != 0

    def on_save_callback(self, content_fp):
        """Return True to continue finding covers, False to stop finding
        covers."""

        self.current += 1
        if self.max_images > 1:
            path = self.path.replace("<imagenum>", str(self.current))
        else:
            path = self.path

        with open(path, 'wb') as fp:
            shutil.copyfileobj(content_fp, fp)

        if self.max_images > 1:
            # XXX: progress_cb makes sense only if we are downloading several
            # images, since it is supposed to update the choose artwork
            # dialog...
            return self.progress_cb(path, self.current-1)

    def on_err_cb(self, reason=None):
        """Return True to stop finding, False to continue finding covers."""
        return False


class ArtworkCache:
    def __init__(self, config, path=None):
        self.logger = logging.getLogger('sonata.artwork.cache')
        self._cache = {}
        self.config = config
        self.path = path if path is not None else \
                os.path.expanduser("~/.config/sonata/art_cache")

    def set(self, key, value):
        self.logger.debug("Setting %r to %r", key, value)
        self._cache[key] = value

    def get(self, key):
        self.logger.debug("Requesting for %r", key)
        return self._cache.get(key)

    def get_pixbuf(self, key, size, default=None):
        self.logger.debug("Requesting pixbuf for %r", key)
        try:
            path = self._cache[key]
        except KeyError:
            return default

        if not os.path.exists(path):
            self._cache.pop(key, None)
            return default

        try:
            p = GdkPixbuf.Pixbuf.new_from_file_at_size(path, size, size)
        except:
            self.logger.exception("Unable to load %r at size (%d, %d)",
                                  path, size, size)
            raise
        return img.do_style_cover(self.config, p, size, size)

    def save(self):
        self.logger.debug("Saving to %s", self.path)
        misc.create_dir(os.path.dirname(self.path))
        try:
            with open(self.path, 'w', encoding="utf8") as f:
                f.write(repr(self._cache))
        except IOError as e:
            self.logger.info("Unable to save: %s", e)

    def load(self):
        self.logger.debug("Loading from %s", self.path)
        self._cache = {}
        try:
            with open(self.path, 'r', encoding="utf8") as f:
                self._cache = eval(f.read())
        except (IOError, SyntaxError) as e:
            self.logger.info("Unable to load: %s", e)
